//! Machine-level Intermediate Representation

use std::collections::btree_map::Entry;
use std::collections::{BTreeMap, BTreeSet, HashMap, HashSet, VecDeque};

use itertools::Itertools;
use petgraph::visit::{IntoEdges, NodeRef};

use crate::string_table::{Index as StringsIndex, StringTable};

#[derive(Debug, PartialEq, Eq, Clone, Copy, Hash)]
pub enum Type {
    /// 8bit integer
    Byte,
    /// 16bit integer
    Word,
    /// 32bit integer
    DWord,
    /// 64bit integer
    QWord,
    /// 32bit ieee-754
    SinglePrecision,
    /// 64bit ieee-754
    DoublePrecision,
    // XWord,
    // YWord,
    // ZWord,
}

impl core::fmt::Display for Type {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let name = match self {
            Type::Byte => "byte",
            Type::Word => "word",
            Type::DWord => "dword",
            Type::QWord => "qword",
            Type::SinglePrecision => "f32",
            Type::DoublePrecision => "f64",
        };

        write!(f, "{name}")
    }
}

impl Type {
    pub const fn is_floating(self) -> bool {
        match self {
            Type::SinglePrecision | Type::DoublePrecision => true,
            _ => false,
        }
    }
    pub const fn from_bitsize_int(size: u32) -> Type {
        match size {
            ..=8 => Self::Byte,
            9..=16 => Self::Word,
            17..=32 => Self::DWord,
            33..=64 => Self::QWord,
            _ => unimplemented!(),
        }
    }
    pub const fn from_bytesize_int(size: u32) -> Type {
        Self::from_bitsize_int(size * 8)
    }

    pub const fn bits(self) -> u32 {
        match self {
            Type::Byte => 8,
            Type::Word => 16,
            Type::DWord => 32,
            Type::QWord => 64,
            Type::SinglePrecision => 32,
            Type::DoublePrecision => 64,
        }
    }

    pub const fn constant_inst(self) -> Inst {
        match self {
            Type::Byte => Inst::ConstantByte,
            Type::Word => Inst::ConstantWord,
            Type::SinglePrecision | Type::DWord => Inst::ConstantDWord,
            Type::DoublePrecision | Type::QWord => Inst::ConstantQWord,
        }
    }

    pub const fn bytes(self) -> u32 {
        match self {
            Type::Byte => 1,
            Type::Word => 2,
            Type::DWord => 4,
            Type::QWord => 8,
            Type::SinglePrecision => 4,
            Type::DoublePrecision => 8,
        }
    }
}

#[derive(Debug, PartialEq, Eq, Clone, Copy, Hash)]
pub enum Inst {
    /// index
    Label,
    // Constant(Type),
    /// index
    ConstantBytes,
    /// imm8
    ConstantByte,
    /// imm16
    ConstantWord,
    /// imm32
    ConstantDWord,
    /// imm64
    ConstantQWord,
    /// src
    LoadConstant(Type), // hint for loading constant into register
    /// ast-node
    ExternRef,
    /// size, align
    Alloca,
    /// src
    Load(Type),
    /// src, dst
    Store(Type),
    /// src, idx,
    GetElementPtr(Type),
    Parameter(Type),
    /// lhs, rhs
    Add(Type),
    /// lhs, rhs
    Sub(Type),
    /// lhs, rhs
    Mul(Type),
    /// lhs, rhs
    MulSigned(Type),
    /// lhs, rhs
    Div(Type),
    /// lhs, rhs
    DivSigned(Type),
    /// lhs, rhs
    Rem(Type),
    /// lhs, rhs
    RemSigned(Type),
    /// lhs, rhs
    BitAnd(Type),
    /// lhs, rhs
    BitOr(Type),
    /// lhs, rhs
    BitXOr(Type),
    /// lhs
    Negate(Type),
    /// lhs, rhs
    ShiftLeft(Type),
    /// lhs, rhs
    ShiftRightSigned(Type),
    /// lhs, rhs
    ShiftRightUnsigned(Type),
    /// lhs
    ReturnValue,
    Return,
    //FunctionStart,
}

impl Inst {
    fn value_type(&self) -> Option<Type> {
        match self {
            Inst::Label
            | Inst::ConstantBytes
            | Inst::ConstantByte
            | Inst::ConstantWord
            | Inst::ConstantDWord
            | Inst::ConstantQWord
            | Inst::ExternRef
            | Inst::Alloca
            | Inst::ReturnValue
            | Inst::Store(_)
            | Inst::Return => None,
            Inst::GetElementPtr(ty)
            | Inst::Load(ty)
            | Inst::LoadConstant(ty)
            | Inst::Parameter(ty)
            | Inst::Add(ty)
            | Inst::Sub(ty)
            | Inst::Mul(ty)
            | Inst::MulSigned(ty)
            | Inst::Div(ty)
            | Inst::DivSigned(ty)
            | Inst::Rem(ty)
            | Inst::RemSigned(ty)
            | Inst::BitAnd(ty)
            | Inst::BitOr(ty)
            | Inst::BitXOr(ty)
            | Inst::Negate(ty)
            | Inst::ShiftLeft(ty)
            | Inst::ShiftRightSigned(ty)
            | Inst::ShiftRightUnsigned(ty) => Some(*ty),
        }
    }
    fn has_value(&self) -> bool {
        // basically, when an arithmetic instruction has two immediates, then just replace it with a mov into the dst reg
        match self {
            Inst::Label
            | Inst::ConstantBytes
            | Inst::ConstantByte
            | Inst::ConstantWord
            | Inst::ConstantDWord
            | Inst::ConstantQWord
            | Inst::ExternRef
            | Inst::Alloca
            | Inst::Store(_)
            | Inst::ReturnValue
            | Inst::Return => false,
            Inst::GetElementPtr(_)
            | Inst::Load(_)
            | Inst::LoadConstant(_)
            | Inst::Parameter(_)
            | Inst::Add(_)
            | Inst::Sub(_)
            | Inst::Mul(_)
            | Inst::MulSigned(_)
            | Inst::Div(_)
            | Inst::DivSigned(_)
            | Inst::Rem(_)
            | Inst::RemSigned(_)
            | Inst::BitAnd(_)
            | Inst::BitOr(_)
            | Inst::BitXOr(_)
            | Inst::Negate(_)
            | Inst::ShiftLeft(_)
            | Inst::ShiftRightSigned(_)
            | Inst::ShiftRightUnsigned(_) => true,
        }
    }
}

#[derive(Clone, Copy)]
pub union Data {
    none: (),
    imm8: u8,
    imm16: u16,
    imm32: u32,
    imm64: u64,
    index: StringsIndex,
    node: u32,
    binary: (u32, u32),
}

impl From<u8> for Data {
    fn from(value: u8) -> Self {
        Data::imm8(value)
    }
}
impl From<u16> for Data {
    fn from(value: u16) -> Self {
        Data::imm16(value)
    }
}
impl From<u32> for Data {
    fn from(value: u32) -> Self {
        Data::imm32(value)
    }
}
impl From<u64> for Data {
    fn from(value: u64) -> Self {
        Data::imm64(value)
    }
}

impl Data {
    pub fn imm8(v: u8) -> Data {
        Self { imm8: v }
    }
    pub fn imm16(v: u16) -> Data {
        Self { imm16: v }
    }
    pub fn imm32(v: u32) -> Data {
        Self { imm32: v }
    }
    pub fn imm64(v: u64) -> Data {
        Self { imm64: v }
    }
    pub fn index(v: StringsIndex) -> Data {
        Self { index: v }
    }
    pub fn node(v: u32) -> Data {
        Self { node: v }
    }
    pub fn binary(lhs: u32, rhs: u32) -> Data {
        Self { binary: (lhs, rhs) }
    }
    pub fn none() -> Data {
        Self { none: () }
    }

    pub fn as_imm8(self) -> u8 {
        unsafe { self.imm8 }
    }
    pub fn as_imm16(self) -> u16 {
        unsafe { self.imm16 }
    }
    pub fn as_imm32(self) -> u32 {
        unsafe { self.imm32 }
    }
    pub fn as_imm64(self) -> u64 {
        unsafe { self.imm64 }
    }
    pub fn as_index(self) -> StringsIndex {
        unsafe { self.index }
    }
    pub fn as_node(self) -> u32 {
        unsafe { self.node }
    }
    pub fn as_binary(self) -> (u32, u32) {
        unsafe { self.binary }
    }
}

pub struct Mir {
    pub nodes: Vec<Inst>,
    pub data: Vec<Data>,
}

impl Mir {
    pub fn new() -> Mir {
        Self {
            nodes: Vec::new(),
            data: Vec::new(),
        }
    }

    pub fn push(&mut self, inst: Inst, data: Data) -> u32 {
        let node = self.nodes.len() as u32;
        self.nodes.push(inst);
        self.data.push(data);

        node
    }

    pub fn gen_u8(&mut self, value: u8) -> u32 {
        self.push(Inst::ConstantByte, Data::imm8(value))
    }
    pub fn gen_u16(&mut self, value: u16) -> u32 {
        self.push(Inst::ConstantWord, Data::imm16(value))
    }
    pub fn gen_u32(&mut self, value: u32) -> u32 {
        self.push(Inst::ConstantDWord, Data::imm32(value))
    }
    pub fn gen_u64(&mut self, value: u64) -> u32 {
        self.push(Inst::ConstantQWord, Data::imm64(value))
    }
    pub fn gen_load_const(&mut self, ty: Type, src: u32) -> u32 {
        self.push(Inst::LoadConstant(ty), Data::node(src))
    }

    pub fn gen_label(&mut self, name: StringsIndex) -> u32 {
        self.push(Inst::Label, Data::index(name))
    }
    pub fn gen_alloca(&mut self, size: u32, align: u32) -> u32 {
        self.push(Inst::Alloca, Data::binary(size, align))
    }
    pub fn gen_load(&mut self, ty: Type, src: u32) -> u32 {
        self.push(Inst::Load(ty), Data::node(src))
    }
    pub fn gen_get_element_ptr(&mut self, ty: Type, src: u32, index: u32) -> u32 {
        self.push(Inst::GetElementPtr(ty), Data::binary(src, index))
    }
    pub fn gen_store(&mut self, ty: Type, src: u32, dst: u32) -> u32 {
        self.push(Inst::Store(ty), Data::binary(src, dst))
    }
    pub fn gen_param(&mut self, ty: Type) -> u32 {
        self.push(Inst::Parameter(ty), Data::none())
    }

    pub fn gen_truncate(&mut self, src: u32, ty: Type, signed: bool, bits: u16) -> u32 {
        // example: u4
        // ty = byte
        // 1111_1111 << 4 = 1111_0000
        // mask = 0000_1111;
        // shift = 8 - 4;
        let shift = ty.bits() as u8 - bits as u8;
        let mask = !(!0u64 << bits);

        let mask = match ty {
            Type::Byte => self.gen_u8(mask as u8),
            Type::Word => self.gen_u16(mask as u16),
            Type::SinglePrecision | Type::DWord => self.gen_u32(mask as u32),
            Type::DoublePrecision | Type::QWord => self.gen_u64(mask as u64),
        };

        let masked = self.gen_bitand(ty, src, mask);

        if signed {
            let shift = self.gen_u8(shift);
            let tmp = self.gen_shl(ty, masked, shift);
            let tmp = self.gen_sar(ty, tmp, shift);
            tmp
        } else {
            masked
        }
    }

    pub fn gen_add(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::Add(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_sub(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::Sub(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_mul(&mut self, ty: Type, signed: bool, lhs: u32, rhs: u32) -> u32 {
        if signed && !ty.is_floating() {
            self.gen_mul_signed(ty, lhs, rhs)
        } else {
            self.gen_mul_unsigned(ty, lhs, rhs)
        }
    }
    pub fn gen_mul_unsigned(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::Mul(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_mul_signed(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::MulSigned(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_div(&mut self, ty: Type, signed: bool, lhs: u32, rhs: u32) -> u32 {
        if signed && !ty.is_floating() {
            self.gen_div_signed(ty, lhs, rhs)
        } else {
            self.gen_div_unsigned(ty, lhs, rhs)
        }
    }
    pub fn gen_div_unsigned(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::Div(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_div_signed(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::DivSigned(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_rem(&mut self, ty: Type, signed: bool, lhs: u32, rhs: u32) -> u32 {
        if signed && !ty.is_floating() {
            self.gen_rem_signed(ty, lhs, rhs)
        } else {
            self.gen_rem_unsigned(ty, lhs, rhs)
        }
    }
    pub fn gen_rem_unsigned(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::Rem(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_rem_signed(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::RemSigned(ty), Data::binary(lhs, rhs))
    }

    pub fn gen_bitand(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::BitAnd(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_bitor(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::BitAnd(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_bitxor(&mut self, ty: Type, lhs: u32, rhs: u32) -> u32 {
        self.push(Inst::BitAnd(ty), Data::binary(lhs, rhs))
    }
    pub fn gen_negate(&mut self, ty: Type, src: u32) -> u32 {
        self.push(Inst::Negate(ty), Data::node(src))
    }

    #[doc(alias = "gen_shift_left")]
    pub fn gen_shl(&mut self, ty: Type, src: u32, shift: u32) -> u32 {
        self.push(Inst::ShiftLeft(ty), Data::binary(src, shift))
    }
    #[doc(alias = "gen_shift_right")]
    pub fn gen_shr(&mut self, ty: Type, src: u32, shift: u32) -> u32 {
        self.push(Inst::ShiftRightUnsigned(ty), Data::binary(src, shift))
    }
    #[doc(alias = "gen_shift_right_signed")]
    pub fn gen_sar(&mut self, ty: Type, src: u32, shift: u32) -> u32 {
        self.push(Inst::ShiftRightSigned(ty), Data::binary(src, shift))
    }
    pub fn gen_ret_val(&mut self, val: u32) -> u32 {
        self.push(Inst::ReturnValue, Data::node(val))
    }
    pub fn gen_ret(&mut self) -> u32 {
        self.push(Inst::Return, Data::none())
    }
}

impl Mir {
    fn render_node<W: core::fmt::Write>(
        &self,
        w: &mut W,
        strings: &StringTable,
        reg_alloc: &BTreeMap<u32, amd64::Register>,
        node: u32,
    ) -> core::fmt::Result {
        let idx = node as usize;
        let inst = self.nodes[idx];
        let data = self.data[idx];

        if let Some(reg) = reg_alloc.get(&node) {
            write!(w, "({reg}) ")?;
        }

        match inst {
            Inst::Label => writeln!(w, "%{node} = {}", strings.get_str(data.as_index())),
            Inst::ConstantBytes => writeln!(
                w,
                "%{node} = bytes({:x?})",
                strings.get_bytes(data.as_index())
            ),
            Inst::ConstantByte => writeln!(w, "%{node} = imm8({:x?})", data.as_imm8()),
            Inst::ConstantWord => writeln!(w, "%{node} = imm16({:x?})", data.as_imm16()),
            Inst::ConstantDWord => writeln!(w, "%{node} = imm32({:x?})", data.as_imm32()),
            Inst::ConstantQWord => writeln!(w, "%{node} = imm64({:x?})", data.as_imm64()),
            Inst::LoadConstant(ty) => {
                let src = data.as_node();
                writeln!(w, "%{node} = load constant {ty} %{src}")
            }
            Inst::ExternRef => writeln!(w, "%{node} = extern %%{}", data.as_node()),
            Inst::Alloca => {
                let (size, align) = data.as_binary();
                writeln!(w, "%{node} = alloca {size}, {align}")
            }
            Inst::GetElementPtr(ty) => {
                let (ptr, idx) = data.as_binary();
                writeln!(w, "%{node} = get element ptr {ty}, ptr %{ptr}, idx {idx}")
            }
            Inst::Load(ty) => {
                writeln!(w, "%{node} = load {ty}, ptr %{}", data.as_node())
            }
            Inst::Store(ty) => {
                let (src, dst) = data.as_binary();
                writeln!(w, "%{node} = store {ty}, ptr %{dst}, {ty} %{src}")
            }
            Inst::Parameter(ty) => {
                writeln!(w, "%{node} = param {ty}")
            }
            Inst::Add(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = add {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Sub(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = sub {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Mul(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = mul {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::MulSigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = signed mul {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Div(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = div {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::DivSigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = signed div {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Rem(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = rem {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::RemSigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = signed rem {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::BitAnd(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = bitand {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::BitOr(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = bitor {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::BitXOr(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = bitxor {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::Negate(ty) => {
                let lhs = data.as_node();
                writeln!(w, "%{node} = negate {ty} %{lhs}")
            }
            Inst::ShiftLeft(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = shift left {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::ShiftRightUnsigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = shift right {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::ShiftRightSigned(ty) => {
                let (lhs, rhs) = data.as_binary();
                writeln!(w, "%{node} = signed shift right {ty} %{lhs}, {ty} %{rhs}")
            }
            Inst::ReturnValue => {
                let lhs = data.as_node();
                writeln!(w, "%{node} = return %{lhs}")
            }
            Inst::Return => {
                writeln!(w, "%{node} = return")
            }
        }
    }

    pub fn render<W: core::fmt::Write>(
        &self,
        w: &mut W,
        strings: &StringTable,
    ) -> core::fmt::Result {
        let reg_alloc = self.build_liveness();
        for node in 0..self.nodes.len() {
            self.render_node(w, strings, &reg_alloc, node as u32)?;
        }
        Ok(())
    }

    pub fn display<'a, 'b>(&'a self, strings: &'b StringTable) -> DisplayMir<'a, 'b> {
        DisplayMir { mir: self, strings }
    }
}

impl Mir {
    pub fn build_liveness(&self) -> BTreeMap<u32, amd64::Register> {
        struct Interval {
            start: u32,
            end: u32,
        }

        let mut references = BTreeMap::<u32, Vec<u32>>::new();
        let mut preferred_colors = BTreeMap::<u32, amd64::Register>::new();
        use amd64::Register::*;
        let mut in_colors = [r9, r8, rcx, rdx, rdi, rsi].to_vec();
        let mut in_colors_sse = [xmm7, xmm6, xmm5, xmm4, xmm3, xmm2, xmm1, xmm0].to_vec();
        let mut inouts = Vec::<u32>::new();

        for i in 0..self.nodes.len() {
            let inst = self.nodes[i];
            let data = self.data[i];

            let node = i as u32;
            references.insert(node, Vec::new());

            match inst {
                Inst::Parameter(ty) => {
                    if let Some(reg) = if ty.is_floating() {
                        in_colors_sse.pop()
                    } else {
                        in_colors.pop()
                    } {
                        println!("prefering {reg} for param");
                        preferred_colors.insert(node, reg);
                    };
                    inouts.push(node);
                }
                // return is thru rax.
                Inst::ReturnValue => {
                    let val = data.as_node();
                    inouts.push(val);
                    references.get_mut(&val).unwrap().push(node);
                    _ = preferred_colors.try_insert(val, amd64::Register::rax);
                }
                Inst::Negate(_) | Inst::Load(_) => {
                    references.get_mut(&data.as_node()).unwrap().push(node);
                }
                Inst::GetElementPtr(_) => {
                    let (src, _) = data.as_binary();
                    references.get_mut(&src).unwrap().push(node);
                }
                // mul wants lhs (or rhs) to be rax, imul doesn't care as much.
                // note that it also clobers rdx
                Inst::Mul(ty) | Inst::MulSigned(ty) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);

                    if !ty.is_floating() {
                        _ = preferred_colors.try_insert(lhs, amd64::Register::rax);
                        _ = preferred_colors.try_insert(rhs, amd64::Register::rax);
                    }
                }
                // div wants lhs to be rax, idiv can do either.
                // note that it also clobers rdx
                Inst::Div(ty) | Inst::DivSigned(ty) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);

                    if !ty.is_floating() {
                        _ = preferred_colors.try_insert(lhs, amd64::Register::rax);
                    }
                }
                // div wants lhs to be rax, idiv can do either.
                // note that it also clobers rax
                Inst::Rem(ty) | Inst::RemSigned(ty) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);

                    if !ty.is_floating() {
                        _ = preferred_colors.try_insert(lhs, amd64::Register::rdx);
                    }
                }
                // shr,shl,sar want the shift to be in cl.
                Inst::ShiftLeft(_) | Inst::ShiftRightSigned(_) | Inst::ShiftRightUnsigned(_) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);
                    _ = preferred_colors.try_insert(rhs, amd64::Register::rcx);
                }
                // add,adc,sub,sbb,or,and,xor and mov don't care much about their source registers
                Inst::Add(_)
                | Inst::Sub(_)
                | Inst::BitAnd(_)
                | Inst::BitOr(_)
                | Inst::BitXOr(_)
                | Inst::Store(_) => {
                    let (lhs, rhs) = data.as_binary();
                    references.get_mut(&lhs).unwrap().push(node);
                    references.get_mut(&rhs).unwrap().push(node);
                }
                _ => {}
            }
        }

        references.retain(|&node, refs| !refs.is_empty() && self.nodes[node as usize].has_value());
        preferred_colors.retain(|&node, _| self.nodes[node as usize].has_value());

        let intervals = references
            .iter()
            .map(|(node, refs)| {
                (
                    *node,
                    Interval {
                        start: *node,
                        end: refs.last().cloned().unwrap(),
                    },
                )
            })
            .collect::<BTreeMap<u32, Interval>>();

        let mut edges = HashSet::<(u32, u32)>::new();
        for (&node, interval) in &intervals {
            for (&other, _) in references.range(interval.start + 1..interval.end) {
                edges.insert((node, other));
            }
        }
        let inference_graph = petgraph::graph::UnGraph::<(), ()>::from_edges(edges.into_iter());

        #[derive(Debug, Default, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
        enum Color {
            #[default]
            Unassigned,
            Tentative(amd64::Register),
            Final(amd64::Register),
        }

        impl Color {
            fn color(self) -> Option<amd64::Register> {
                match self {
                    Color::Unassigned => None,
                    Color::Tentative(color) | Color::Final(color) => Some(color),
                }
            }
        }

        struct Colorizer<'a> {
            mir: &'a Mir,
            graph: petgraph::graph::UnGraph<(), ()>,
            colors: BTreeMap<u32, Color>,
            preferred: BTreeMap<u32, amd64::Register>,
        }

        impl<'a> Colorizer<'a> {
            fn node_colors(&self, node: petgraph::graph::NodeIndex) -> &[amd64::Register] {
                let colors = if self.mir.nodes[node.index()]
                    .value_type()
                    .map(|t| t.is_floating())
                    == Some(true)
                {
                    &amd64::Register::SSE[..]
                } else {
                    &amd64::Register::GPR[..]
                };
                colors
            }

            fn prepass<I: IntoIterator<Item = u32>>(&mut self, inouts: I) {
                for node in inouts.into_iter() {
                    self.precolor_node(node.into());
                }

                let keys = self.preferred.keys().cloned().collect::<Vec<_>>();
                for node in keys {
                    self.precolor_node(node.into());
                }
            }

            fn precolor_node(&mut self, node: petgraph::graph::NodeIndex) {
                // prepass: assign preferred colours for in/out values and specific
                // instructions like mul/div which require one operand to be in rax
                let node_u32 = node.index() as u32;
                // only apply color here if we have a preference
                if let Some(preferred_color) = self.preferred.remove(&node_u32) {
                    let mut clique_colors = self
                        .graph
                        .neighbors(node)
                        .filter_map(|n| self.colors.get(&(n.index() as u32)).cloned());

                    if clique_colors
                        .find(|color| color.color() == Some(preferred_color))
                        .is_none()
                    {
                        self.colors
                            .insert(node_u32, Color::Tentative(preferred_color));
                    }
                };
                // .chain(self.node_colors(node).into_iter().cloned());
            }

            fn color_node(&mut self, node: petgraph::graph::NodeIndex) {
                // final pass:
                // look at clique colors and prefer to steal colors from
                // tentatively colored nodes. this results in preferential
                // coloring depending on the order of the prepass.
                let node_u32 = node.index() as u32;
                let clique_colors = self
                    .graph
                    .neighbors(node)
                    .filter_map(|n| self.colors.get(&(n.index() as u32)).cloned())
                    .collect::<BTreeSet<_>>();

                let colors = self
                    .node_colors(node)
                    .into_iter()
                    .filter(|&&r| !clique_colors.contains(&Color::Final(r)))
                    .cloned()
                    .collect::<BTreeSet<_>>();

                // eprintln!("coloring %{node_u32}:");
                // eprintln!("\twants: {:?}", self.colors.get(&node_u32));
                // eprintln!("\tclique: {clique_colors:?}");
                // eprintln!("\tcandidates: {colors:?}");

                match self.colors.entry(node_u32) {
                    Entry::Vacant(v) => {
                        // here we want to first check clique_colors with tentative coloring.
                        let color = colors
                            .into_iter()
                            .find_or_first(|&c| !clique_colors.contains(&Color::Tentative(c)))
                            .expect("ran out of registers :(");
                        v.insert(Color::Final(color));
                    }
                    Entry::Occupied(mut e) => {
                        // we prefer to steal
                        variant!(e.get() => &Color::Tentative(reg));
                        let color = colors
                            .into_iter()
                            .find_or_first(|&c| c == reg)
                            .expect("ran out of registers :(");
                        e.insert(Color::Final(color));
                    }
                }
            }

            fn finalise(self) -> BTreeMap<u32, amd64::Register> {
                self.colors
                    .into_iter()
                    .filter_map(|(node, c)| match c {
                        Color::Final(reg) => Some((node, reg)),
                        _ => None,
                    })
                    .collect()
            }
        }

        let mut colorizer = Colorizer {
            mir: self,
            graph: inference_graph,
            preferred: preferred_colors,
            colors: BTreeMap::new(),
        };

        // prepass: assign preferred colours for in/out values and specific
        // instructions like mul/div which require one operand to be in rax
        colorizer.prepass(inouts);

        for &node in references.keys().rev() {
            if !self.nodes[node as usize].has_value() {
                continue;
            }
            colorizer.color_node(node.into());
        }

        let colors = colorizer.finalise();

        // eprintln!(
        //     "Inference Graph:\n{:?}",
        //     petgraph::dot::Dot::with_attr_getters(
        //         &inference_graph,
        //         &[
        //             petgraph::dot::Config::EdgeNoLabel,
        //             petgraph::dot::Config::NodeNoLabel,
        //         ],
        //         &Box::new(|_, _| { "".to_owned() }),
        //         &Box::new(|_, node: (petgraph::graph::NodeIndex, &())| {
        //             let (node, _) = node;
        //             if let Some(reg) = assigned_colors.get(&(node.index() as u32)) {
        //                 format!("label=\"{} ({})\"", node.index(), reg)
        //             } else {
        //                 format!("label=\"{}\"", node.index())
        //             }
        //         }),
        //     ),
        // );

        colors
    }
}

pub struct DisplayMir<'a, 'b> {
    mir: &'a Mir,
    strings: &'b StringTable,
}

impl<'a, 'b> core::fmt::Display for DisplayMir<'a, 'b> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.mir.render(f, &self.strings)
    }
}
